﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using System.Windows.Forms;
using System.IO;
using System.Text.RegularExpressions;
using Ionic.Zip;
using CTNet;

namespace ChkBoxCtrl
{
    public partial class Form1 : Form
    {
        /*
         * Processes the boxes in the 'Set Manipulation' list by checking against each box
         * set in the 'Check Box Sets' list for duplicate boxes. Returns a list of
         * 'Set Manipulation' boxes that are not duplicated in the 'Check Box Sets' list.
         * 
         * The user is informed of duplicates with an error message and the duplicate box(es)
         * are highlighted in RED in both lists.
         * 
         * The global lists dupManipList and dupIndxList are cleared then updated if there are
         * duplicate boxes found in the 'Check Box Sets' setList.
         *
         * Called From:
         *      setManipUpdateButton_Click()
         *      setManipAddAboveButton_Click()
         *      setManipAddBelowButton_Click()
         *      
         * Inputs:
         *      manipTxt: The name of the List and the button that was clicked on to either
         *                  update or add a new Check Box Set.
         *        setTxt: The text that identifies the Check Box Sets we are checking
         *                  for duplicates.
         *      selIndex: The index of the selected set in the 'Check Box Sets' list.
         *                selIndex will be set to -1 when we check all sets in the
         *                'Check Box Sets' list for duplicates. We do this for the
         *                'New Set Above' and the 'New Set Below' operations.  
        */
        public List<string> ProcessDuplicateBoxes(string manipTxt, string setTxt, int selIndex)
        {
            List<string> manipBoxList = new List<string>();
            List<string> boxIntersectList = new List<string>();
            List<string> tmpSetList = new List<string>();

            bool dupsFound = false;

            // Clear these out, we will rebuild them here if any duplicate boxes are found in other box sets.
            dupManipList.Clear();
            dupIndxList.Clear();

            // Store boxes in the 'Set Manipulation' list in our local list.
            manipBoxList = setManipList.Items.OfType<string>().ToList();

            // For each set we will check if we have a duplicate box. For the 'Update Set' operation
            // we will skip over the currently selected set since we are updating it.
            for (int i = 0; i < setList.Items.Count; i++)
            {
                if (i != selIndex)
                {
                    tmpSetList = setList.Items[i].ToString().Replace("[", "").Replace("]", "").Split(',').ToList();
                    boxIntersectList = manipBoxList.Intersect(tmpSetList).ToList();

                    if (boxIntersectList.Count > 0)
                    {
                        dupsFound = true;

                        foreach (string box in boxIntersectList)
                        {
                            if (!dupManipList.Contains(box))
                            {
                                dupManipList.Add(box);
                            }

                            // Will catch any duplicate boxes that are in more than 1 set other
                            // than the selected set. This 'should' not happen.
                            dupIndxList.Add(i);

                            ProcessErrors("E", manipTxt, 3, string.Format("{0} is already present in '{1}'", box, setTxt), null);
                        }
                    }
                }
            }

            // Debug Start.
            //for (int i = 0; i < dupManipList.Count; i++)
            //{
            //    ProcessErrors("W", "ProcessDuplicateBoxes()", 0, string.Format("dupManipList[{0}]: {1}", i, dupManipList[i]), null);
            //}
            //for (int i = 0; i < dupIndxList.Count; i++)
            //{
            //    ProcessErrors("W", "ProcessDuplicateBoxes()", 0, string.Format("dupIndxList[{0}]: {1}", i, dupIndxList[i]), null);
            //}
            // Debug Stop.

            // Refresh the 'Check Box Sets' lists to reflect any duplicate boxes.
            if (dupsFound)
            {
                setList.Refresh();
            }

            /*
             * Doing this outright because there are scenarios where these boxes
             * were RED coming in to this method, but should be reset. An example
             * would be attempting a 'New Set Above' on a selected setList set
             * [we did nothing to the setManipList boxes, they are all the boxes
             * in the selected set], in which case all setManipList boxes would be
             * red because we check ALL setList sets for that operation, then doing an 
             * 'Update Set' operation -> coming here the boxes are red, up above we
             * clear out the dupManipList AND there are no duplicates encountered here
             * due to the 'Update Set' operation does not check against the selected
             * setList set.
            */
            setManipList.Refresh();

            // Return the 'Set Manipulation' boxes minus any 'Check Box Sets' duplicate boxes.
            return manipBoxList.Except(dupManipList).ToList();
        }

        /*
         *
         * This routine displays error/warning messages to the user via MessageBox.Show().
         *
         * Called From:
         *      xmlFilterSetListButton_Click()
         *      ProcessInputCode()
         *      copySetsButton_Click()
         *      
         * Inputs:
         *      errType: Either "E" for Error or "W" for Warning.
         *      titleStr: Used in preparing the string for displaying in the MessageBox title bar.
         *                Will be null if not used.
         *      errNum: The error number which is our index into the string array errMsg.
         *      var1: Caller sets this up, will be null if not used.
         *      var1: Caller sets this up, will be null if not used.
        */
        public void ProcessErrors(string errType, string titleStr, int errNum, string var1, string var2)
        {
            string eType = (errType == "W" ? "Warning" : "Error");
            MessageBoxIcon icon = (errType == "W" ? MessageBoxIcon.Warning : MessageBoxIcon.Error);
            string msgTitle = (titleStr == null ? "" : titleStr + " - ") + eType + "!";

            // Error message 7 - Start
            int num = 0;
            string var3 = "";
            string var4 = "";

            if (int.TryParse(var2, out num) == true)
            {
                var3 = (num > 1 ? "s" : "");
                var4 = (num > 1 ? "ve" : "s");
            }
            // Error message 7 - Stop.

            string[] errMsg = {
                // This slot is for debug messages.
                string.Format("  {0}\n", var1),

                string.Format("  {0} - XML filtering has resulted in {1} single box set(s):\t\n\nThese will need to either be deleted from the '{2}' list or will need more boxes added before 'Check Box Sets' code can be generated.", eType.ToUpper(), var1, var2),
                string.Format("  {0} - Duplicate Box: {1}\t\n\n  This box is already in the '{2}' list.\n\n  Box will not be added.", eType.ToUpper(), var1, var2),
                string.Format("  {0} - Duplicate Box:\t\n\n  {1}", eType.ToUpper(), var1),
                string.Format("  {0} - Box sets must have more than one box.\n\n  Operation aborted.", eType.ToUpper()),
                string.Format("  {0} - {1} single box set(s) detected while retrieving\n  Check Box Sets:\t\n\nThese will need to either be deleted from the '{2}' list or will need more boxes added before 'Check Box Sets' code can be generated.", eType.ToUpper(), var1, var2),
                string.Format("  {0} - Check Box Sets list is empty:\t\n\n  'Check Box Sets' code can not be generated.", eType.ToUpper()),
                string.Format("  {0} - One Box Set:\t\n\n  The '{1}' list contains {2} set{3} that ha{4} only 1 box.\n  Box sets must have more than one box.\n\n  Operation aborted.", eType.ToUpper(), var1, var2, var3, var4),
                string.Format("  {0} - State Box Incorrect: '{1}'\t\n\n  State Box is either missing or is not formatted correctly.", eType.ToUpper(), var1),
                string.Format("  {0} - launchMode is neither 'editChkBoxSets' nor 'addChkBoxCtrl':\t\n\n  Mode can not be processed correctly.", eType.ToUpper())
                };

            MessageBox.Show(this, errMsg[errNum], msgTitle, MessageBoxButtons.OK, icon);
        }

        /*
        * This uses the OpenFileDialog class in allowing the user to select a file
        * which will be used in retrieving checkboxes from an .XML file.
        *
        * Called From:
        *   zipFileToolStripMenuItem_Click()
        *   xMLFilereadCheckboxesToolStripMenuItem_Click()
        *
        * Inputs:
        *   dir: The initial file path from which a file may be selected.
        *   type: The type of file we will read to extract the boxes from, either "ZIP" or "XML".
        *
        * Returns the full path and file name of the selected file or null if no file was selected.
        */
        public string GetFileName(string dir, string type)
        {
            string fileName = null;
            OpenFileDialog ofd = new OpenFileDialog();

            ofd.InitialDirectory = dir;

            ofd.Filter = (type == "ZIP" ? "ZIP file|*.zip" : "XML file|*.xml");

            System.Windows.Forms.DialogResult dr = ofd.ShowDialog();

            if (dr == DialogResult.OK)
            {
                // This holds the path and the file name with extension.
                fileName = ofd.FileName;
            }
            return fileName;
        }

        /*
         * This extracts the "DOC.XML" file from the ZIP file identified by the argument zipFileName.
         *
         * Called From:
         *  Form1(string[] args)
         *  xMLFiledocumentBoxNumbersAndBoxWidthsToolStripMenuItem_Click()
         *
         * Inputs:
         *  zipFileName: The full path and file name of the zip file.
         *
         * Returns the full path and file name of the extracted "DOC.XML" file or null if the exraction
         *  process did not work out well.
        */
        public string UnzipXmlFile(string zipFileName)
        {
            string docXmlFile = null;
            string tmpPath;
            // Got this password from Joe A.
            string password = "5e$ACre7ru";

            //MessageBox.Show(string.Format("UnzipXmlFile - zipFileName: {0}", zipFileName));

            try
            {
                // We first need to get the temp path where we will store the extracted XML file, the
                // caller will delete the file after reading it for checkbox information.
                tmpPath = Path.GetTempPath();

                if (!String.IsNullOrEmpty(tmpPath))
                {
                    //MessageBox.Show(string.Format("UnzipXmlFile - tmpPath: {0}", tmpPath));

                    using (ZipFile zip = ZipFile.Read(zipFileName))
                    {
                        foreach (ZipEntry e in zip)
                        {
                            if (e.FileName.ToLower() == "doc.xml")
                            {
                                if (e.UsesEncryption)
                                {
                                    e.Password = password;
                                }
                                e.Extract(tmpPath, ExtractExistingFileAction.OverwriteSilently);

                                // Set the path and file name of the extracted XML file to be returned to caller.
                                docXmlFile = tmpPath + "doc.xml";
                            }
                        }
                    }
                }
            }
            catch (System.Security.SecurityException ex)
            {
                MessageBox.Show(ex.ToString(), "Could not get the path to the Temp folder.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }

            return docXmlFile;
        }

        /*
         * Called From:
         *      Form1(string[] args)
         *      zipFileToolStripMenuItem_Click()
         *      xMLFilereadCheckboxesToolStripMenuItem_Click()
         *
         * Inputs:
         *      filename:   The path and filename of the XML file we will inspect for checkboxes.
         *      
         * Globals Affected: 
         * 
         *  boxPageDict - build the dictionary of all checkboxes with their page numbers.
         *  boxListFormat - reflects the box format style in the 'XML Check Boxes' list.
         *  
         * We build the boxListDict with the Key as the XML checkbox,
         * with the Value being the page number for the checkbox. We store
         * this way as the new box format is now devoid of a page number, and
         * it seems appropriate to display boxes in a sorted [by page number]
         * manner for ease in working with the checkboxes on a certain document
         * page.
         * 
         * We then clear the 'XML Check Boxes' list, boxList. And populate the
         * list with all discovered checkboxes.
         * 
         * We also determine the box format the XML boxes are following,
         * new format (box12) or old format (box1_12).
         * 
         * IF we find that the box format styles between the 'XML Check Boxes'
         * and 'Check Box Sets' lists differ we will inform the user and
         * perform the deleteSetListButton_Click() event.
        */
        public void SetXMLCheckBoxList(string fileName)
        {
            bool boxFormatIsSet = false;
            int pageNum = 0;
            string boxKey = null;
            string line;
            Match match;

            try
            {
                boxPageDict.Clear();

                using (StreamReader sr = new StreamReader(fileName))
                {
                    while ((line = sr.ReadLine()) != null)
                    {
                        // Remove whitespace:
                        line = Regex.Replace(line, @"\s+", "");

                        // Catch the current page number.
                        match = Regex.Match(line, @"<pageref=""pg(\d+)"">");

                        if (match.Success)
                        {
                            int.TryParse(match.Groups[1].Value, out pageNum);
                        }
                        else
                        {
                            // Looking for checkboxes:
                            match = Regex.Match(line, @"<textref=""(box\d+_\d+)"".*checkbox=""true""");

                            if (match.Success)
                            {
                                if (boxFormatIsSet == false)
                                {
                                    boxListFormat = boxFormat.oldFormat;
                                    boxFormatIsSet = true;
                                }

                                boxKey = match.Groups[1].Value;
                            }
                            else
                            {
                                match = Regex.Match(line, @"<textref=""(box\d+)"".*checkbox=""true""");

                                if (match.Success)
                                {
                                    if (boxFormatIsSet == false)
                                    {
                                        boxListFormat = boxFormat.newFormat;
                                        boxFormatIsSet = true;
                                    }

                                    boxKey = match.Groups[1].Value;
                                }
                            }

                            if (boxKey != null)
                            {
                                if (!boxPageDict.ContainsKey(boxKey))
                                {
                                    boxPageDict.Add(boxKey, pageNum);
                                }

                                boxKey = null;
                            }
                        }
                    }
                }

                /*
                 * If we have boxes in the dictionary we will insert them
                 * into the boxList sorted on page number.
                 */
                if (boxPageDict.Count > 0)
                {
                    foreach (KeyValuePair<string, int> kvp in boxPageDict.OrderBy(key => key.Value))
                    {
                        boxList.Items.Add(kvp.Key);
                    }

                    if ((setListFormat != boxFormat.notSet) && (boxListFormat != setListFormat))
                    {
                        // Can't have 2 formats at same time:
                        string errMsg;
                        string setListFormatStr;
                        string boxListFormatStr;

                        setListFormatStr = (setListFormat == boxFormat.newFormat ? "new box format" : "old box format");
                        boxListFormatStr = (boxListFormat == boxFormat.newFormat ? "new box format" : "old box format");

                        errMsg = string.Format("  Box Format Error:\n\n  '{0}' list boxes are using {1}.\n\n  '{2}' list boxes are using {3}.\n\n  '{2}' list will be emptied.", boxListLabel.Text, boxListFormatStr, setListLabel.Text, setListFormatStr);

                        MessageBox.Show(errMsg, "Methods.cs-SetXMLCheckBoxList():", MessageBoxButtons.OK, MessageBoxIcon.Error);

                        deleteSetListButton.PerformClick();
                    }

                    /*
                     * Another idiot check: Warn the user if the XML filename,
                     * minus the path and extension does not match the launch
                     * file name (again minus the path and extension) IF this
                     * application was started from the code editor.
                     */
                    if ((editorLaunch == true) && (xmlIdLabel.Text != vscFileInfo.fileNameNoExtTmp))
                    {
                        string errMsg;

                        errMsg = string.Format("  The '{0}' list file name\n\n  {1}\n\n  differs from code editor launch file name\n\n  {2}", boxListLabel.Text, xmlIdLabel.Text, vscFileInfo.fileNameNoExtTmp);

                        MessageBox.Show(errMsg, "Methods.cs-SetXMLCheckBoxList():", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                    }
                }
            }
            catch (IOException ex)
            {
                MessageBox.Show(ex.ToString(), "File could not be read.", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString());
            }
        }

        /*
         * Called From:
         *      ProcessInputCode()
         *      setList_Click()
         *      xmlFilterSetListButton_Click()
         *      boxSelFromCHBtn_Click()
         *      ProcessCHBoxSelOnOff()
         *      
         * First clear the 'Set Manipulation' list, setManipList. If we have a set selected
         * in the 'Check Box Sets' list, setList, we will display those boxes in the setManipList
         * and select the first box.
        */
        public void ProcessSetListSelection()
        {
            MatchCollection matches;

            setManipList.Items.Clear();

            try
            {
                if (setList.Items.Count > 0)
                {
                    if (setList.SelectedItem.ToString() != null)
                    {
                        //MessageBox.Show(string.Format("setListFormat = {0}", setListFormat));

                        // We use Regex.Matches because we expect many boxes as .Matches returns a MatchCollection.
                        if (setListFormat == boxFormat.newFormat)
                        {
                            matches = Regex.Matches(setList.SelectedItem.ToString(), @"box\d+");
                        }
                        else
                        {
                            matches = Regex.Matches(setList.SelectedItem.ToString(), @"box\d+_\d+");
                        }

                        // Iterate over the collection of matches:
                        if (matches.Count > 0)
                        {
                            foreach (Match match in matches)
                            {
                                //MessageBox.Show(string.Format("match = {0}", match.Value));

                                setManipList.Items.Add(match.Value);
                            }

                            //MessageBox.Show(string.Format("setManipList.Count = {0}", setManipList.Items.Count));

                            setManipList.SetSelected(0, true);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessSetListSelection:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * Called From:
         *      Form1()
         *      copyFromClipboardButton_Click()
         *
         * Globals Affected:
         *      dupIndxList - cleared if we find Checkbox sets in global var chkBoxCtrlStr.
         *      dupManipList - cleared if we find Checkbox sets in global var chkBoxCtrlStr.
         *      oneBoxIndxList - cleared if we find Checkbox sets in global var chkBoxCtrlStr.
         *      launchMode - modified if we launch from a code editor.
         *      setListFormat - reflects what box format is being used in the 'Check Box Sets'.
         *      
         * Argument setMode indicates if we are to set the launchMode of the application.
         * We set this when launching from an editor and let the user decide to make any
         * further setting changes. When started as a stand alone app the launchMode is
         * disabled.
         * 
         * This routine searches thru global chkBoxCtrlStr for the set of check boxes we are
         * currently controlling. It is not an error if we do not detect the box sets, we are
         * just looking thru the text supplied by the user.
         * 
         * Identified Checkbox sets will be inserted into the 'Check Box Sets' .ListBox, setList,
         * after we clear it out.
         * 
         * We will also determine what the box format style is being used if we find boxes.
         * This box format style pertains to the 'Check Box Sets' list. If the discovered
         * box format is different from the 'XML Check Boxes' list then we have a problem
         * as there can not be 2 different formats at the same time. We will delete the
         * 'XML Check Boxes' list and notify the user.
        */
        public void ProcessInputCode(bool setMode)
        {
            List<string> boxSetsList = new List<string>();
            List<string> tmpList = new List<string>();
            string tmpStr;
            Match match;
            bool formatSet = true;

            try
            {
                /*
                 * Checking to see if chkBoxCtrlStr contains the Check Box Sets,
                 * if it does, we will display the box sets in the 'Check Box Sets'
                 * list and, if we are in setMode mode also set the launchMode to
                 * 'Edit Check Box Sets' else we will assume we launched in
                 * 'Add Check Box Control' mode.
                 */
                tmpStr = Regex.Replace(chkBoxCtrlStr, "[\n\r\t ]", "");

                match = Regex.Match(tmpStr, @"return\[(\[""box\d+_\d+""(,""box\d+_\d+"")*\](,\[""box\d+_\d+""(,""box\d+_\d+"")*\])*)\];");
                if (match.Success)
                {
                    setListFormat = boxFormat.oldFormat;
                }
                else
                {
                    match = Regex.Match(tmpStr, @"return\[(\[""box\d+""(,""box\d+"")*\](,\[""box\d+""(,""box\d+"")*\])*)\];");

                    if (match.Success)
                    {
                        setListFormat = boxFormat.newFormat;
                    }
                    else
                    {
                        /*
                         * Leaving setListFormat as it is which also means we are
                         * leaving setList and related lists as is.
                         */
                        formatSet = false;
                    }
                }

                if (formatSet == true)
                {
                    /*
                     * Using '!' as a placeholder so we can split the string into individual box
                     * sets for displaying in the setList.
                     */
                    boxSetsList = match.Groups[1].Value.Replace("\"", "").Replace(",[", "![").Split('!').ToList();

                    if (modeToolStripMenuItem.Enabled && setMode)
                    {
                        launchMode = LaunchMode.editChkBoxSets;
                    }
                }
                else
                {
                    if (modeToolStripMenuItem.Enabled && setMode)
                    {
                        launchMode = LaunchMode.addChkBoxCtrl;
                    }
                }

                if (boxSetsList.Count > 0)
                {
                    // Clear these out to get ready for the new box sets:
                    dupIndxList.Clear();
                    dupManipList.Clear();
                    oneBoxIndxList.Clear();

                    setList.Items.Clear();
                    setList.HorizontalExtent = 0;

                    // Insert the box sets into the cleared Set List.
                    for (int i = 0; i < boxSetsList.Count; i++)
                    {
                        // Remove any duplicate boxes [should not be any, but...] from each set.
                        tmpList = boxSetsList[i].Replace("[", "").Replace("]", "").Split(',').Distinct().ToList();

                        if (tmpList.Count == 1)
                        {
                            // Will warn user about 1 box sets we encounter here.
                            oneBoxIndxList.Add(i);
                        }

                        //MessageBox.Show(string.Format("ProcessInputCode(): setList.Items.Add({0})", "[" + string.Join(",", tmpList.ToArray()) + "]"));

                        setList.Items.Add("[" + string.Join(",", tmpList.ToArray()) + "]");
                    }

                    /*
                     * This will run thru the setList and set the 
                     * horizontal scrollbar if needed.
                     */
                    SetListHorExtent();

                    if (oneBoxIndxList.Count > 0)
                    {
                        // We either read Check Box Control code from the Clipboard or from the code editor file.
                        ProcessErrors("W", (editorLaunch ? "Copied From Editor" : copyFromClipboardButton.Text), 5, oneBoxIndxList.Count.ToString(), setListLabel.Text);
                    }

                    // Select the first set:
                    setList.SetSelected(0, true);

                    // Display first set contents in set_manip_list.
                    ProcessSetListSelection();

                    if ((boxListFormat != boxFormat.notSet) && (boxListFormat != setListFormat))
                    {
                        // Can't have 2 formats at same time:
                        string errMsg;
                        string setListFormatStr;
                        string boxListFormatStr;

                        setListFormatStr = (setListFormat == boxFormat.newFormat ? "new box format" : "old box format");
                        boxListFormatStr = (boxListFormat == boxFormat.newFormat ? "new box format" : "old box format");

                        errMsg = string.Format("  Box Format Error:\n\n  '{0}' list boxes are using {1}\n  '{2}' list boxes are using {3}\n\n  '{2}' list will be emptied.", setListLabel.Text, setListFormatStr, boxListLabel.Text, boxListFormatStr);

                        MessageBox.Show(errMsg, "Methods.cs-ProcessInputCode():", MessageBoxButtons.OK, MessageBoxIcon.Error);

                        boxList.Items.Clear();

                        boxListFormat = boxFormat.notSet;
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessInputCode:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * Called From:
         *      Form1.cs-ctClient_CtClientEvent()
         *
         * Globals Affected:
         *      dupIndxList - cleared if we find Checkbox sets in arg boxSetsStr.
         *      dupManipList - cleared if we find Checkbox sets in arg boxSetsStr.
         *      oneBoxIndxList - cleared if we find Checkbox sets in arg boxSetsStr.
         *      setListFormat - reflects what box format is being used in the 'Check Box Sets'.
         *      
         * This routine searches thru arg boxSetsStr for the set of check boxes we are
         * controlling in Calc Helper.
         * 
         * Identified Checkbox sets will be inserted into the 'Check Box Sets'
         * .ListBox, setList, after we clear it out.
         * 
         * We directly set the setList box format as Calc Helper and Checkbox Control
         * are tied to the same code file.
         * 
         * We also adjust the Launch Mode if required - User may just launch Checkbox
         * Control with no sets in the clipboard knowing that they will get checkbox
         * selections from Calc Helper which sends the filtered sets over.
        */
        public void ProcessCHBoxSets(string boxSetsStr)
        {
            //MessageBox.Show(string.Format("Entered ProcessCHBoxSets(): boxSetsStr == {0}\nboxSetsStr.Length == {1}", boxSetsStr, boxSetsStr.Length));

            List<string> boxSetsList = new List<string>();
            List<string> tmpList = new List<string>();
            Match match;

            // Clear out to get ready for the new box sets:
            dupIndxList.Clear();
            dupManipList.Clear();
            oneBoxIndxList.Clear();

            setList.Items.Clear();
            setList.HorizontalExtent = 0;

            // Right now there is no selected set.
            boxSetsIndex = -1;

            try
            {
                /*
                 * We directly set the setList box format since we got this
                 * box set string from Calc Helper and both tools are linked
                 * to the exact code file.
                 * 
                 * Calc Helper has used a '!' to separate box sets in the
                 * string so we look for them here instead of ','. Note that
                 * there still will be commas inside a set between boxes.
                 */
                match = Regex.Match(boxSetsStr, @"(\[box\d+_\d+(,box\d+_\d+)*\](!\[box\d+_\d+(,box\d+_\d+)*\])*)");

                if (match.Success)
                {
                    setListFormat = boxFormat.oldFormat;
                }
                else
                {
                    match = Regex.Match(boxSetsStr, @"(\[box\d+(,box\d+)*\](!\[box\d+(,box\d+)*\])*)");

                    if (match.Success)
                    {
                        setListFormat = boxFormat.newFormat;
                    }
                    else
                    {
                        setListFormat = boxFormat.notSet;
                    }
                }

                boxSetsList = boxSetsStr.Split(new string[] { "!" }, StringSplitOptions.RemoveEmptyEntries).ToList();

                //MessageBox.Show(string.Format("In ProcessCHBoxSets(): boxSetsList.Count == {0}", boxSetsList.Count));

                if (boxSetsList.Count > 0)
                {
                    // Insert the box sets into the cleared Set List.
                    for (int i = 0; i < boxSetsList.Count; i++)
                    {
                        // Check for a single box set - will warn the user:
                        match = Regex.Match(boxSetsList[i], @"^\[box\d+(_\d+)?\]$");

                        if (match.Success)
                        {
                            oneBoxIndxList.Add(i);
                        }

                        setList.Items.Add(boxSetsList[i]);
                    }

                    /*
                     * Set the horizontal scrollbar if required:
                     */
                    SetListHorExtent();

                    if (oneBoxIndxList.Count > 0)
                    {
                        ProcessErrors("W", "Box Sets From Calc Helper", 5, oneBoxIndxList.Count.ToString(), setListLabel.Text);
                    }

                    if (modeToolStripMenuItem.Enabled)
                    {
                        if (launchMode != LaunchMode.editChkBoxSets)
                        {
                            ProcessMode(LaunchMode.editChkBoxSets, false);
                        }
                    }
                }
                else
                {
                    if (modeToolStripMenuItem.Enabled)
                    {
                        if (launchMode != LaunchMode.addChkBoxCtrl)
                        {
                            ProcessMode(LaunchMode.addChkBoxCtrl, false);
                        }
                    }
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessCHBoxSets:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }

            //MessageBox.Show(string.Format("Leaving ProcessCHBoxSets(): setList.Items.Count == {0}", setList.Items.Count));
        }

        /*
         * Calc Helper has sent over a new checkbox selection index.
         * 
         * Called From:
         * 
         *      Form1.cs-ctClient_CtClientEvent()
         */
        public void ProcessNewCHSelIndex(int selIndex)
        {
            //MessageBox.Show(string.Format("Entered ProcessNewCHSelIndex():\n\nselIndex = {0}\nboxSetsIndex = {1}", selIndex, boxSetsIndex));

            try
            {
                boxSetsIndex = selIndex;
                setList.SelectedIndex = selIndex;

                /*
                 * Allow the user to say they are done editing the currently
                 * selected set and want to start a new set.
                 */
                xmlFilterSetListButton.Enabled = true;
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessNewCHSelIndex:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * When called from deleteSetListSelButton_Click() AND Calc Helper
         * is NOT sending over checkbox selections:
         * 
         * Deletes the 'Check Box Sets' current selection and selects a new
         * 'Check Box Set' and displays the boxes in the new selection in
         * the 'Set Manipulation' list. setList only allows one set at a
         * time to be selected.
         * 
         * When called from deleteSetListSelButton_Click() AND we are
         * receiving boxes from CalcHelper:
         * 
         * Send Calc Helper a 'delBoxSet' message along with the deleted
         * set's setList index.
         * 
         * When NOT called from deleteSetListSelButton_Click():
         * 
         * The senderId would be "CH". The user removed the last box
         * from the set currently being edited.
         * 
         * Called From:
         * 
         *      Form1.cs-ctClient_CtClientEvent()-ctEvent.boxSel
         *      Form1.cs-deleteSetListSelButton_Click()
         */
        public void ProcessSetListSetDeletion(string senderId, int currIndex)
        {
            //MessageBox.Show(string.Format("Entered ProcessSetListSetDeletion():\nsenderId = {0}\ncurrIndex = {1}\nboxSetsIndex = {2}", senderId, currIndex, boxSetsIndex));

            int lastIndex = setList.Items.Count - 1;

            try
            {
                /*
                 * Check if the selection is a one box set, if so, we need to
                 * remove the selected index from oneBoxIndxList.
                 */
                if (oneBoxIndxList.Contains(currIndex))
                {
                    oneBoxIndxList.Remove(currIndex);
                }

                /*
                 * We now adjust the oneBoxIndxList values [if any] to reflect
                 * the selected set removal from this list as indexes will need
                 * decrementing depending on the selected set position in the list.
                 * The 'XML Filter' operation is the only source of one box sets.
                 * The user is not allowed to make one via 'Update Box Set' nor
                 * 'New Box Set'.
                 */
                for (int i = 0; i < oneBoxIndxList.Count; i++)
                {
                    if (currIndex < oneBoxIndxList[i])
                    {
                        oneBoxIndxList[i] -= 1;
                    }
                }

                // Now the deletion.
                setList.Items.RemoveAt(currIndex);

                //MessageBox.Show(this, string.Format("senderId = {0}\nsetList.SelectedIndex = {1}\ncurrIndex = {2}\nboxSetsIndex = {3}", senderId, setList.SelectedIndex, currIndex, boxSetsIndex));

                /*
                 * Reset the .HorizontalExtent if we deleted the last
                 * set in the list. This will remove the horizontal scrollbar
                 * if present [list DrawMode is OwnerDrawFixed].
                 */
                if (setList.Items.Count == 0)
                {
                    setList.HorizontalExtent = 0;
                    setListFormat = boxFormat.notSet;
                }
                else
                {
                    /*
                     * Set the horizontal scrollbar if required:
                     */
                    SetListHorExtent();
                }

                if ((ctClient != null) && ctClient.IsConnected() && (ctClient.boxSelFromCH == true))
                {
                    if (boxSetsIndex == currIndex)
                    {
                        xmlFilterSetListButton.Enabled = false;
                        boxSetsIndex = -1;
                    }

                    if (senderId != "CH")
                    {
                        // Inform Calc Helper about the deletion:
                        boxSelMsgTypes msgType = boxSelMsgTypes.delBoxSet;

                        ctClient.SendSelectedIndexMsg(string.Format("{0}|{1}", msgType, currIndex));
                    }
                }
                else
                {
                    /*
                     * Clear the dupIndxList and the dupManipList since we are
                     * deleting the set they were associated with.
                     */
                    dupIndxList.Clear();
                    dupManipList.Clear();

                    /*
                     * Now select either at the same currIndex or at the 'new'
                     * lastIndex if we deleted the last set in the list.
                     */
                    if (lastIndex > 0)
                    {
                        setList.SetSelected((currIndex < lastIndex ? currIndex : lastIndex - 1), true);
                    }

                    /*
                     * Display the selected set's boxes in the 'Set Manipulation'
                     * list if we have a selection, else we just clear that list.
                     */
                    ProcessSetListSelection();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessSetListSetDeletion:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This replaces the box set at the current setList
         * selection with the updated version from Calc Helper.
         * 
         * Called From:
         * 
         *      ctClient_CtClientEvent()-ctEvent.boxSel
         */
        public void ProcessCHBoxSetUpdate(string boxSet)
        {
            //MessageBox.Show(string.Format("Entered ProcessCHBoxSetUpdate():\n\n boxSetsIndex: {0}\n box set:\n{1}", boxSetsIndex, boxSet));

            try
            {
                List<string> boxList = boxSet.Replace("[", "").Replace("]", "").Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                if (boxList.Count > 0)
                {
                    if (boxList.Count > 1)
                    {
                        if (oneBoxIndxList.Contains(boxSetsIndex) == true)
                        {
                            oneBoxIndxList.Remove(boxSetsIndex);
                        }
                    }
                    else
                    {
                        if (oneBoxIndxList.Contains(boxSetsIndex) == false)
                        {
                            oneBoxIndxList.Add(boxSetsIndex);
                        }
                    }

                    setList.Items[boxSetsIndex] = boxSet;

                    /*
                     * To scrollbar or not to scrollbar, that is
                     * the horizontal question.
                     */
                    SetListHorExtent();
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessCHBoxSetUpdate:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This adds the new checkbox set at the end of
         * the 'Check Box Sets' list and adjusts the
         * boxSetsIndex. We then select the new set.
         * 
         * We also set the setListFormat if not set already.
         * This could happen, if Checkbox Control starts up
         * from the editor and there are no current
         * checkboxes on the document and the user requests
         * checkbox selections from Calc Helper.
         * 
         * Called From:
         * 
         *      ctClient_CtClientEvent()-ctEvent.newCHSet
         */
        public void ProcessNewCHBoxSet(string newSet)
        {
            try
            {
                Match match;
                List<string> boxList = newSet.Replace("[", "").Replace("]", "").Split(new string[] { "," }, StringSplitOptions.RemoveEmptyEntries).ToList();

                //MessageBox.Show(string.Format("Entered ProcessNewCHBoxSet():\n\n boxSetsIndex: {0}\n new set:\n{1}", boxSetsIndex, newSet));

                if (boxList.Count > 0)
                {
                    if (setListFormat == boxFormat.notSet)
                    {
                        // It will suffice to use the first box in the list.
                        match = Regex.Match(boxList[0], @"^box\d+_\d+$");

                        if (match.Success)
                        {
                            setListFormat = boxFormat.oldFormat;
                        }
                        else
                        {
                            match = Regex.Match(boxList[0], @"^box\d+$");

                            if (match.Success)
                            {
                                setListFormat = boxFormat.newFormat;
                            }
                        }
                    }

                    boxSetsIndex = setList.Items.Count;

                    if (boxList.Count > 1)
                    {
                        if (oneBoxIndxList.Contains(boxSetsIndex) == true)
                        {
                            oneBoxIndxList.Remove(boxSetsIndex);
                        }
                    }
                    else
                    {
                        if (oneBoxIndxList.Contains(boxSetsIndex) == false)
                        {
                            oneBoxIndxList.Add(boxSetsIndex);
                        }
                    }

                    setList.Items.Add(newSet);

                    /*
                     * Set the horizontal scrollbar if required:
                     */
                    SetListHorExtent();

                    setList.SelectedIndex = boxSetsIndex;

                    /*
                     * Allow the user to indicate that they want to
                     * start a new set.
                     */
                    xmlFilterSetListButton.Enabled = true;
                }
            }
            catch (Exception ex)
            {
                MessageBox.Show(ex.ToString(), "ProcessNewCHBoxSet:", MessageBoxButtons.OK, MessageBoxIcon.Error);
            }
        }

        /*
         * This sets setList.HorizontalExtent property
         * so we can see the horizontal scrollbar.
         * This is because this ListBox has the DrawMode
         * property set to Owner Drawn.
         * 
         * The longest set in the list sets the length.
         * Might end up not having a horizontal scrollbar.
         * Friends tell me horizontal is the way to go.
         * 
         * Called From:
         * 
         *  ProcessSetListSetDeletion()
         *  ProcessCHBoxSetUpdate()
         *  setManipUpdateButton_Click()
         *  ProcessInputCode()
         *  xmlFilterSetListButton_Click()
         *  setManipAddAboveButton_Click()
         *  setManipAddBelowButton_Click()
         *  ProcessNewCHBoxSet()
         *  ProcessCHBoxSets()
        */
        public void SetListHorExtent()
        {
            Graphics g = setList.CreateGraphics();
            int hzSize;

            setList.HorizontalExtent = 0;

            foreach (string set in setList.Items)
            {
                hzSize = (int)(g.MeasureString(set, setList.Font).Width + 0.5f) + 25;

                if (hzSize > setList.HorizontalExtent)
                {
                    setList.HorizontalExtent = hzSize;
                }
            }
        }

        /*
         * This sets the mode and message label text [top area of the form] and
         * also resets the Mode vars copySetsClicked and copyMethClicked.
         * We use the Mode vars to determine when to automatically exit the app
         * after a copy to clipboard button has been clicked. We also disable the
         * copyMethCallButton if changing to 'Edit Check Box Sets' mode.
         * 
         * We only deal with launchMode when we have a successfule launch from the
         * code editor.
         * 
         * The red background has been set in the form constructor to indicate a
         * editor launch of this application.
         * 
         * Called From:
         *  Form1() - constructor at startup from an editor.
         * addCheckBoxControlToolStripMenuItem_Click()
         * editCheckBoxSetsToolStripMenuItem_Click()
         * 
         * Inputs:
         *  newLaunchMode  The enum that tells us the launchMode. Note that this can
         *      also be changed by the user at any time before the application closes.
        */
        public void ProcessMode(LaunchMode newLaunchMode, bool startup)
        {
            if (!startup)
            {
                if (launchMode == newLaunchMode)
                {
                    return;
                }

                copySetsClicked = false;
                copyMethClicked = false;
                copyMethCallButton.BackColor = Color.LightGray;
                copySetsButton.BackColor = Color.LightGray;

                launchMode = newLaunchMode;
            }

            if (newLaunchMode == LaunchMode.editChkBoxSets)
            {
                modeLabel.Text = "Mode: Edit Check Box Sets";
                msgLabel.Text = "STARTED FROM A CODE EDITOR - APPLICATION WILL EXIT AFTER YOU CLICK ON THE 'Copy Check Box Sets To Clipboard' BUTTON BELOW";

                copyMethCallButton.Enabled = false;
            }
            else
            {
                modeLabel.Text = "Mode: Add Check Box Control";
                msgLabel.Text = "STARTED FROM A CODE EDITOR - APPLICATION WILL EXIT AFTER YOU CLICK ON BOTH COPY TO CLIPBOARD BUTTONS BELOW";

                copyMethCallButton.Enabled = true;
            }
        }

        /*
         * Handles processing when we are either about to get
         * checkbox selections from Calc Helper or we are no
         * longer going to get them.
         * 
         * Called By:
         * 
         *      Form1.cs-ctClient_CtClientEvent()
         *      Form1.cs-boxSelFromCHBtn_Click()
         */
        public void ProcessCHBoxSelOnOff(bool chBoxSelOn)
        {
            if (chBoxSelOn == true)
            {
                /*
                 * ctEvent that brought us here:
                 * 
                 * ctEvent.boxSelAck
                 */
                boxSelFromCHBtn.BackgroundImage = ChkBoxCtrl.Properties.Resources.BoxSelFromCHOn;
                boxSelFromCHBtn.ToolTipText = "CalcHelper Is Now In 'Checkbox Control Tool' Box Selection Mode";

                // Clear the setManipList:
                setManipDelListButton.PerformClick();

                /*
                 * ********************
                 * These controls will be disabled as they are not not needed.
                 */
                readToolStripMenuItem.Enabled = false;

                // boxList controls:
                panel4.Enabled = false;

                // Set Manipulation controls:
                panel5.Enabled = false;

                copyFromClipboardButton.Enabled = false;

                /*
                 * ********************
                 */

                // Change of functionality for this button:
                xmlFilterSetListButton.Text = "New Set";

                /*
                 * We enable this when the user is editing a box
                 * set, i.e. a set is currently selected in setList.
                 */
                xmlFilterSetListButton.Enabled = false;

                ctClient.BoxSetsRequest();
            }
            else
            {
                /*
                 * ctEvent that brought us here:
                 * 
                 * ctEvent.docLockOrNoCheckboxes
                 * ctEvent.disconnectCH
                 * ctEvent.boxSelOff
                 */
                boxSelFromCHBtn.BackgroundImage = ChkBoxCtrl.Properties.Resources.BoxSelFromCHOff;
                boxSelFromCHBtn.ToolTipText = "Click To Receive Checkbox Selections From CalcHelper";

                // Back to life:
                readToolStripMenuItem.Enabled = true;

                // boxList controls:
                panel4.Enabled = true;

                // Set Manipulation controls:
                panel5.Enabled = true;

                xmlFilterSetListButton.Text = "XML Filter";
                xmlFilterSetListButton.Enabled = true;

                copyFromClipboardButton.Enabled = true;

                /*
                 * Putting these back to the Checkbox Control
                 * launch appearance.
                 */
                if (setList.Items.Count > 0)
                {
                    setList.SetSelected(0, true);
                    ProcessSetListSelection();
                }
            }
        }

        /*
         * Send current setList checkbox sets to Calc Helper
         * so it can update it's checkbox control state list.
         * This is just in case the user does not update the
         * real checkbox control sets in the visual studio
         * code file. Want to keep the 'visual' checkbox control
         * appearance on Calc Helper consistent with the real
         * checkbox control when a refresh operation is performed
         * on Calc Helper.
         * 
         * Called From:
         * 
         * Form1.cs-Form1_FormClosing()
         * Form1.cs-boxSelFromCHBtn_Click()
         */
        public void SendSetStateMsg()
        {
            string msgStr = "";
            boxSelMsgTypes msgType = boxSelMsgTypes.setCBCState;

            if (setList.Items.Count > 0)
            {
                foreach (string set in setList.Items)
                {
                    msgStr += (set.Replace("[", "").Replace("]", "") + ",");
                }

                msgStr = msgStr.TrimEnd(',');
            }

            ctClient.SendSelectedIndexMsg(string.Format("{0}|{1}", msgType, msgStr));
        }
    }
}